/*
 * Copyright (c) 2015 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/
/** API usually used in libdbus/libglib.
 \defgroup API API
 */
#ifndef _LIBDBUSPOLICY1_H_
#define _LIBDBUSPOLICY1_H_

#include <stdarg.h>
#include <sys/types.h>

#ifdef __cplusplus
extern "C" {
#endif

#define SYSTEM_BUS_CONF_FILE_PRIMARY  "/usr/share/dbus-1/system.conf"
#define SESSION_BUS_CONF_FILE_PRIMARY "/usr/share/dbus-1/session.conf"

#define SYSTEM_BUS_CONF_FILE_SECONDARY  "/etc/dbus-1/system.conf"
#define SESSION_BUS_CONF_FILE_SECONDARY "/etc/dbus-1/session.conf"

/** used when check policy for message prepared to send */
#define DBUSPOLICY_DIRECTION_SENDING 0

 /** used when check policy for message read from bus */
#define DBUSPOLICY_DIRECTION_RECEIVING 1

#define DBUSPOLICY_MESSAGE_TYPE_METHOD_CALL     1
#define DBUSPOLICY_MESSAGE_TYPE_METHOD_RETURN   2
#define DBUSPOLICY_MESSAGE_TYPE_ERROR           3
#define DBUSPOLICY_MESSAGE_TYPE_SIGNAL          4
#define DBUSPOLICY_RESULT_ALLOW                 1
#define DBUSPOLICY_RESULT_DENY                  0
#define DBUSPOLICY_RESULT_DEST_NOT_AVAILABLE   -1
#define DBUSPOLICY_RESULT_KDBUS_ERROR          -2
#define DBUSPOLICY_RESULT_CYNARA_ERROR         -3

struct udesc;

/**
 * Initialize libdbuspolicy configuration context
 * \param[in] bus_path path to the kdbus bus (system or session)

 * \note This function should be called only on well known kdbus buses
 * - the system bus (/sys/fs/kdbus/0-system/bus) and session bus
 * (/sys/fs/kdbus/getuid()-user/bus).  If any other bus is specified
 * function will not succeed.

 * \return On success pointer to configuration context is returned.  On
 * error NULL is returned.
 * \ingroup API
 *
 * \deprecated This is old-style API function, which forces creating of new
 * bus connection by the library. Use dbuspolicy1_init_shared() and
 * dbuspolicy1_init_set_pool() instead.
 */
void* dbuspolicy1_init(const char *bus_path) __attribute__ ((deprecated));

/**
 * This is a new-style initialization function, which allows sharing kdbus connection with
 * libdbuspolicy client.
 *
 * Instead of single function, new API provides two functions. The dbuspolicy1_init_shared()
 * function initializes the library with bus path and open file descriptor, used by client
 * to connect to kdbus, and returns handle to the configuration, or NULL if client is not allowed
 * to connect to the bus. The dbuspolicy1_init_set_pool() function completes the initialization
 * by setting pointer to the pool.
 * The process is as follows:
 * - client opens a bus file;
 * - client passes the path and the file descriptor to libdbuspolicy;
 * - libdbuspolicy checks if the client is allowed to connect to the bus,
 *   returns handle to the configuration;
 * - client connects to the bus (if allowed);
 * - client maps the kdbus pool;
 * - client completes libdbuspolicy initialization by passing the pointer to the pool.
 * \ingroup API
 * \param[in] bus_path path to the kdbus bus (system or session)
 * \param[in] fd file descriptor to the path
 */
void* dbuspolicy1_init_shared(const char *bus_path, int fd);

/**
 * Completes shared initialization by setting pointer to the pool.
 *
 * \ingroup API
 * \param[in] configuration Pointer with policy configuration acquired using dbuspolicy1_init_shared
 * \param[in] pool Pointer to the pool.
 */
void dbuspolicy1_init_set_pool(void *configuration, void *pool);

/**
 * Free libdbuspolicy configuration context
 * \ingroup API
 * \param[in] configuration Pointer with policy configuration
 */
void dbuspolicy1_free(void* configuration);

/**
 * Check policy for outgoing message
 * \ingroup API
 * \param[in] configuration Pointer with policy configuration
 * \param[in] destination List of message destinations
 * \param[in] sender List of message sender names
 * \param[in] path Path
 * \param[in] interface Interface name
 * \param[in] member Member name
 * \param[in] message_type Message type
 * \param[in] error_name (future implementation)
 * \param[in] reply_serial (future implementation)
 * \param[in] requested_reply (future implementation)
 * \return Allow/deny or negative error code
 */
int dbuspolicy1_check_out(void* configuration,
						  const char *destination,
						  const char *sender,
						  const char *path,
						  const char *interface,
						  const char *member,
						  int         message_type,
						  const char *error_name,
						  int         reply_serial,
						  int         requested_reply);

/** Check policy for incoming message
 * \ingroup API
 * \param[in] configuration Pointer with policy configuration
 * \param[in] destination List of message destinations
 * \param[in] sender List of message sender names
 * \param[in] sender_label Sender label (should be manually extracted from incoming message)
 * \param[in] sender_uid Sender uid (should be manually extracted from incoming message)
 * \param[in] sender_gid Sender gid (should be manually extracted from incoming message)
 * \param[in] path Path
 * \param[in] interface Interface name
 * \param[in] member Member name
 * \param[in] message_type Message type
 * \param[in] error_name (future implementation)
 * \param[in] reply_serial (future implementation)
 * \param[in] requested_reply (future implementation)
 * \returns Allow/deny or negative error code
 */
int dbuspolicy1_check_in(void* configuration,
						 const char *destination,
						 const char *sender,
						 const char *sender_label,
						 uid_t       sender_uid,
						 gid_t       sender_gid,
						 const char *path,
						 const char *interface,
						 const char *member,
						 int         message_type,
						 const char *error_name,
						 int         reply_serial,
						 int         requested_reply);

/**
 * Check policy for service ownership
 * \ingroup API
 * \param[in] configuration Pointer with policy configuration
 * \param[in] service Service name
 * \returns Allow/deny or negative error code
 */
int dbuspolicy1_can_own(void* configuration, const char* const service);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif
