/*
 * Copyright (c) 2016-2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/
/**
 * \file
 * \ingroup Implementation
 */
#include "policy.hpp"
#include "storage_backend_xml.hpp"
#include "tslog.hpp"
#include <boost/algorithm/string/predicate.hpp>
#include <cassert>
#include <cctype>
#include <cstdlib>
#include <cstring>
#include <map>
#include <sys/types.h>

using namespace ldp_xml_parser;

size_t ldp_xml_parser::get_string_heap_allocated_memory(const std::string& x) {
	/* an empty string will have nothing allocated on the heap;
	* it can still have some capacity regardless due to short string optimisation,
	* in which case strings up to this length shouldn't allocate anything on the heap either */
	static const std::string empty_str;
	if (x.capacity() == empty_str.capacity())
		return 0;
	else
		return x.capacity();
}

DecisionItem::DecisionItem(Decision decision, const char* privilege)
	: __decision(decision)
{
	setPrivilege(privilege);
}

void DecisionItem::setPrivilege(const char* privilege)
{
	if (privilege == nullptr) {
		__privilege.clear();
	} else {
		__privilege = privilege;
	}

}

Decision DecisionItem::getDecision() const {
	return __decision;
}

const std::string &DecisionItem::getPrivilege() const {
	return __privilege;
}

size_t DecisionItem::getSize() const {
	return get_string_heap_allocated_memory(__privilege);
}

ItemOwn::ItemOwn(const char* name,
				 Decision decision,
				 const char* privilege)
	:   __decision(DecisionItem(decision, privilege)), __is_prefix(false) {
	setName(name);
}

void ItemOwn::setName(const char* name) {
	if (name == nullptr) {
		__name.clear();
	} else {
		__name = name;
	}
}

const std::string &ItemOwn::getName() const {
	return __name;
}

bool ItemOwn::isPrefix() const {
	return __is_prefix;
}

bool ItemOwn::isMatchAll() const {
	return __name.empty();
}

const DecisionItem& ItemOwn::getDecision() const {
	return __decision;
}

MatchItemSR::MatchItemSR(const char* i,
						 const char* me,
						 const char* p,
						 MessageType t,
						 const KdbusBusNames &n_array)
	: names{n_array}
	, type{t}
{
	if (i) interface = i;
	if (me) member = me;
	if (p) path = p;
}

bool MatchItemSR::match(MessageType _type,
						const boost::string_ref &_interface,
						const boost::string_ref &_path,
						const boost::string_ref &_member,
						const boost::string_ref &_name,
						bool _is_name_prefix,
						Decision decision) const {
	if (_type != MessageType::ANY && _type != type)
		return false;

	auto equal_or_empty = [] (const boost::string_ref &lhs, const boost::string_ref &rhs) {
		return lhs.empty() || rhs.empty() || lhs == rhs;
	};

	/* A special clause for no-interface messages. If an allow rule is for some interface
	 * and message has no interface we have to ignore the rule.
	 * See dbus/bus/policy.c:bus_client_policy_check_can_send() for reference.
	 */
	if (interface.empty() && !_interface.empty() && Decision::DENY != decision)
		return false;

	if (!equal_or_empty(_interface, interface))
		return false;

	if (!equal_or_empty(_path, path))
		return false;

	if (!equal_or_empty(_member, member))
		return false;

	if (!_name.empty()) {
		for (const auto &match_name: names) {
			if (!boost::algorithm::starts_with(match_name, _name))
				continue;

			if (match_name.length() == _name.length())
				return true;
			else if (_is_name_prefix && match_name[_name.length()] == '.')
				return true;
		}
		return false;
	}

	return true;
}

ItemSendReceive::ItemSendReceive()
	: 	__type(MessageType::ANY),
		__is_name_prefix(false) {
}

const DecisionItem& ItemSendReceive::getDecision() const {
	return __decision;
}

const std::string &ItemSendReceive::getName()  const {
	return __name;
}

const std::string &ItemSendReceive::getInterface() const {
	return __interface;
}

const std::string &ItemSendReceive::getMember() const {
	return __member;
}

const std::string &ItemSendReceive::getPath() const {
	return __path;
}

MessageType ItemSendReceive::getType() const {
	return __type;
}

bool ItemSendReceive::isNamePrefix() const {
	return __is_name_prefix;
}



size_t ItemSendReceive::getSize() const
{
	size_t size = __decision.getSize();
	if (!__interface.empty())
		size += __interface.length() + 1;
	if (!__member.empty())
		size += __member.length() + 1;
	if (!__path.empty())
		size += __path.length() + 1;
	return size;
}

const DecisionItem& ItemAccess::getDecision() const {
	return __decision;
}

void ItemAccess::setUser(uid_t user) {
	__type = BusAccessType::USER;
	__uid = user;
}

void ItemAccess::setGroup(gid_t group) {
	__type = BusAccessType::GROUP;
	__gid = group;
}

size_t ItemAccess::getSize() const
{
	return __decision.getSize();
}

uid_t ItemAccess::getUid() const {
	return __uid;
}

gid_t ItemAccess::getGid() const {
	return __gid;
}

BusAccessType ItemAccess::getType() const {
	return __type;
}


MatchItemAccess::MatchItemAccess(const uid_t uid, const std::vector<gid_t> &gids)
	:	__uid(uid),
		__gids(gids) {
}

uid_t MatchItemAccess::getUid() const
{
	return __uid;
}

const std::vector<gid_t>& MatchItemAccess::getGids() const
{
	return __gids;
}

bool MatchItemAccess::match(BusAccessType type, uid_t uid, gid_t gid) const {
	switch(type) {
		case BusAccessType::ALL_GROUPS:
		case BusAccessType::ALL_USERS:
			return true;
		case BusAccessType::GROUP:
			for (const auto _gid : __gids) {
				if (gid == _gid) {
					return true;
				}
			}
			return false;
		case BusAccessType::USER:
			return uid == __uid;
	}
	return false;
}

void ItemBuilder::setOwnItem() {
	if (__current_item_type != ItemType::OWN) {
		__current_own = ItemOwn();
		__current_item_type = ItemType::OWN;
	}
}

ItemOwn *ItemBuilder::getOwnItem() {
	return &__current_own;
}

void ItemBuilder::setSendItem() {
	if (__current_item_type != ItemType::SEND) {
		__current_send = ItemSend();
		__current_item_type = ItemType::SEND;
	}
}

void ItemBuilder::setReceiveItem() {
	if (__current_item_type != ItemType::RECEIVE) {
		__current_receive = ItemReceive();
		__current_item_type = ItemType::RECEIVE;
	}
}

ItemSendReceive *ItemBuilder::getSendReceiveItem() {
	if (__current_item_type == ItemType::SEND)
		return &__current_send;

	assert(__current_item_type == ItemType::RECEIVE);
	return &__current_receive;
}

void ItemBuilder::setAccessItem() {
	if (__current_item_type != ItemType::ACCESS) {
		__current_access = ItemAccess();
		__current_item_type = ItemType::ACCESS;
	}
}

ItemAccess *ItemBuilder::getAccessItem() {
	return &__current_access;
}

ItemBuilder::ItemBuilder(ldp_xml::StorageBackendXML &storage)
	: __storage(storage), __current_item_type(ItemType::GENERIC)
{
	reset();
}

ItemBuilder::~ItemBuilder() {
}

void ItemBuilder::reset() {
	__decision.__decision = Decision::ANY;
	__decision.setPrivilege(nullptr);
	__current_item_type = ItemType::GENERIC;
}

void ItemBuilder::generateItem(PolicyType& policy_type, PolicyTypeValue& policy_type_value) {
	switch (__current_item_type) {
		case ItemType::OWN:
			__current_own.__decision = __decision;
			__storage.addItem(policy_type, policy_type_value, __current_own);
			break;
		case ItemType::SEND:
			__current_send.__decision = __decision;
			__storage.addItem(policy_type, policy_type_value, __current_send);
			break;
		case ItemType::RECEIVE:
			__current_receive.__decision = __decision;
			__storage.addItem(policy_type, policy_type_value, __current_receive);
			break;
		case ItemType::ACCESS:
			__current_access.__decision = __decision;
			__storage.addItem(policy_type, policy_type_value, __current_access);
			break;
		case ItemType::GENERIC:
			// do nothing
			// this is for rules like <allow eavesdrop="true"/>
			// libdbuspolicy does not support that
			break;
	}
	reset();
}

void ItemBuilder::addOwner(const char* owner) {
	setOwnItem();
	getOwnItem()->setName(owner);
}

void ItemBuilder::addName(const char* name, bool prefix) {
	ItemSendReceive* sr = getSendReceiveItem();

	if (name)
		sr->__name = name;
	else
		sr->__name.clear();

	sr->__is_name_prefix = prefix;
}

void ItemBuilder::addInterface(const char* interface) {
	assert(interface);
	ItemSendReceive* sr = getSendReceiveItem();
	sr->__interface = interface;
}

void ItemBuilder::addMember(const char* member) {
	assert(member);
	getSendReceiveItem()->__member = member;
}

void ItemBuilder::addPath(const char* path) {
	assert(path);
	getSendReceiveItem()->__path = path;
}

void ItemBuilder::addMessageType(MessageType type) {
	getSendReceiveItem()->__type = type;
}

void ItemBuilder::setUserAccess(uid_t user) {
	getAccessItem()->setUser(user);
}

void ItemBuilder::setAllUsersAccess() {
	getAccessItem()->setAllUsers();
}

void ItemBuilder::setGroupAccess(gid_t group) {
	getAccessItem()->setGroup(group);
}

void ItemBuilder::setAllGroupsAccess() {
	getAccessItem()->setAllGroups();
}

void ItemBuilder::addPrivilege(const char* privilege) {
	__decision.setPrivilege(privilege);
}

void ItemBuilder::addDecision(Decision decision) {
	__decision.__decision = decision;
}

void ItemBuilder::setPrefix(bool value) {
	assert(__current_item_type == ItemType::OWN);
	getOwnItem()->__is_prefix = value;
}

PolicyTypeValue::PolicyTypeValue() : context(ContextType::DEFAULT) {
}

PolicyTypeValue::PolicyTypeValue(ContextType type) : context(type) {
}

PolicyTypeValue::PolicyTypeValue(uid_t us) : user(us) {
}
