/*
 * Copyright (c) 2016-2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/


/**
 * \file
 * \ingroup Implementation
 */

#ifndef _POLICY_H
#define _POLICY_H

#include "type_list.h"
#include "bus_names_array.hpp"

#include <boost/utility/string_ref.hpp>
#include <string>
#include <vector>

namespace ldp_xml {
	class StorageBackendXML;
}

namespace ldp_xml_parser
{
	/** Type of sent message*/
	enum class MessageType : uint8_t {
		ANY = 0,
		METHOD_CALL,
		METHOD_RETURN,
		ERROR,
		SIGNAL
	};

	/** Item type. (different items have different content of class */
	enum class ItemType : uint8_t {
		GENERIC,
		OWN,
		SEND,
		RECEIVE,
		ACCESS
	};

	/** Policy type, either context, user or group */
	enum class PolicyType : uint8_t {
		NONE,
		CONTEXT,
		USER,
		GROUP
	};

	/** Bus access restriction type, either context, user or group */
	enum class BusAccessType : uint8_t {
		USER,
		GROUP,
		ALL_USERS,
		ALL_GROUPS
	};

	/** Context type for context policy type */
	enum class ContextType : uint8_t {
		DEFAULT,
		MANDATORY,
		MAX
	};

	/** Decision returned by policy checkers. Is parsed later */
	enum class Decision : uint8_t {
		ANY,
		ALLOW,
		DENY,
		CHECK
	};

	/** Final decision of libdbus policy. Either allow, deny or error code */
	enum class DecisionResult : int8_t {
		CYNARA_ERROR = -3,
		DENY = 0,
		ALLOW
	};

	/** Contains value corresponding to each of policy types */
	union PolicyTypeValue {
		PolicyTypeValue();
		PolicyTypeValue(ContextType type);
		PolicyTypeValue(uid_t us);
		ContextType context;
		uid_t user;
		gid_t group;
	};

	size_t get_string_heap_allocated_memory(const std::string& x);

	class ItemBuilder;

	/** Class contains decision and privilege string */
	class DecisionItem {
	private:
		Decision __decision;
		std::string __privilege;
		void setPrivilege(const char* privilege);
	public:
		friend class ItemBuilder;
		DecisionItem(Decision decision = Decision::ANY, const char* privilege = NULL);
		Decision getDecision() const;
		const std::string &getPrivilege() const;
		size_t getSize() const;
	};

	class MatchItemOwn {
	private:
		std::string _name;
	public:
		explicit MatchItemOwn(const char *name) : _name(name) {}
		const std::string &getName() const { return _name; }
	};

	/** Class contains info about ownership policy item */
	class ItemOwn {
	private:
		DecisionItem __decision;
		std::string __name;
		bool __is_prefix;
		void setName(const char* name);
	public:
		typedef MatchItemOwn match_type;

		friend class ItemBuilder;
		explicit ItemOwn(const char* name = NULL,
						 Decision decision = Decision::ANY,
						 const char* privilege = NULL);
		const DecisionItem& getDecision() const;
		const std::string &getName() const;
		bool isPrefix() const;
		bool isMatchAll() const;
	};

	/** Struct which allows to contain multiple connection names then compared in s/r policy checker */
	class MatchItemSR {
		const KdbusBusNames &names;
		boost::string_ref interface;
		boost::string_ref member;
		boost::string_ref path;
		MessageType type;

		friend std::ostream &operator<<(std::ostream& stream, const ldp_xml_parser::MatchItemSR &item);
	public:
		MatchItemSR(const char* i, const char* me, const char* p, MessageType t, const KdbusBusNames &n_array);

		bool match(MessageType _type,
				   const boost::string_ref &_interface,
				   const boost::string_ref &_path,
				   const boost::string_ref &_member,
				   const boost::string_ref &_name,
				   bool _is_name_prefix,
				   Decision decision) const;

		const KdbusBusNames &getNames() const { return names; }
	};

	class MatchItemSend : public MatchItemSR {
		using MatchItemSR::MatchItemSR;
	};

	class MatchItemReceive : public MatchItemSR {
		using MatchItemSR::MatchItemSR;
	};

	/** Class contains info about item send/receive */
	class ItemSendReceive {
	protected:
		DecisionItem __decision;
		std::string __name;
		std::string __interface;
		std::string __member;
		std::string __path;
		MessageType __type;
		bool __is_name_prefix;
		ItemSendReceive();
	public:
		friend class ItemBuilder;
		const DecisionItem& getDecision() const;
		const std::string &getName() const;
		const std::string &getInterface() const;
		const std::string &getMember() const;
		const std::string &getPath() const;
		MessageType getType() const;
		bool isNamePrefix() const;
		size_t getSize() const;
	};

	template <typename M>
	class ItemSR : public ItemSendReceive {
	public:
		typedef M match_type;
	};

	typedef ItemSR<MatchItemSend> ItemSend;
	typedef ItemSR<MatchItemReceive> ItemReceive;

	class MatchItemAccess {
	private:
		const uid_t __uid;
		const std::vector<gid_t> &__gids;

	public:
		MatchItemAccess(const uid_t uid, const std::vector<gid_t> &gids);
		uid_t getUid() const;
		const std::vector<gid_t>& getGids() const;

		bool match(BusAccessType type, uid_t uid, gid_t gid) const;
	};

	/** DBus bus access constraint */
	class ItemAccess {
	private:
		DecisionItem __decision{Decision::ANY};
		uid_t __uid{0};
		gid_t __gid{0};
		BusAccessType __type{BusAccessType::ALL_USERS};

		void setAllUsers() { __type = BusAccessType::ALL_USERS; }
		void setUser(uid_t user);
		void setAllGroups() { __type = BusAccessType::ALL_GROUPS; }
		void setGroup(gid_t group);
	public:
		typedef MatchItemAccess match_type;

		friend class ItemBuilder;
		const DecisionItem& getDecision() const;
		uid_t getUid() const;
		gid_t getGid() const;
		BusAccessType getType() const;
		size_t getSize() const;
	};

	using MatchItemTypes = type_list::type_list<MatchItemOwn, MatchItemSend, MatchItemReceive, MatchItemAccess>;
	template <typename T> using MatchItemIdx = type_list::IDX<T, MatchItemTypes>;

	/** Allows to create items from given name, decision etc. */
	class ItemBuilder {
	private:
		ldp_xml::StorageBackendXML &__storage;

		DecisionItem __decision;
		ItemType __current_item_type;

		ItemOwn __current_own;
		ItemSend __current_send;
		ItemReceive __current_receive;
		ItemAccess __current_access;

		ItemOwn* getOwnItem();
		ItemSendReceive* getSendReceiveItem();
		ItemAccess* getAccessItem();
	public:
		explicit ItemBuilder(ldp_xml::StorageBackendXML &storage);
		~ItemBuilder();
		void generateItem(PolicyType& policy_type, PolicyTypeValue& policy_type_value);
		void reset();
		void setOwnItem();
		void setSendItem();
		void setReceiveItem();
		void setAccessItem();
		void addUser(const char* name);
		void addGroup(const char* name);
		void addOwner(const char* owner);
		void addName(const char* name, bool prefix = false);
		void addInterface(const char* interface);
		void addMember(const char* member);
		void addPath(const char* path);
		void addMessageType(MessageType type);
		void addPrivilege(const char* privilege);
		void addDecision(Decision decision);
		void setPrefix(bool value);
		void setUserAccess(uid_t user);
		void setAllUsersAccess();
		void setGroupAccess(gid_t group);
		void setAllGroupsAccess();
	};
}
#endif
