/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/
#include "policy_containers.hpp"
#include "print_content.hpp"
#include "storage_backend_xml.hpp"
#include "tslog.hpp"
#include "xml_parser.hpp"
#include <cassert>
#include <tuple>

using namespace ldp_xml_parser;

namespace ldp_xml {

/* The following two are building blocks for PolicySet class.
 * They specify if a given PolicySet class supports user and group policies.
 */
/****************** UserAndGroupContext ************************/
template <typename P>
class UserAndGroupContext {
protected:
	/** Map with policies for different users */
	std::map<uid_t, P> user;
	/** Map with policies for different groups */
	std::map<gid_t, P> group;

	template <typename T>
	void addItemUser(uid_t uid, T &item) { user[uid].addItem(item); }

	template <typename T>
	void addItemGroup(gid_t gid, T &item) { group[gid].addItem(item); }

public:
	const std::map<uid_t, P> &getPoliciesUser() const { return user; }
	const std::map<uid_t, P> &getPoliciesGroup() const { return group; }
};

/****************** NoUserAndGroupContext ************************/
class NoUserAndGroupContext {
protected:
	template <typename T>
	void addItemUser(uid_t, T &) { assert(false); }

	template <typename T>
	void addItemGroup(gid_t, T &) { assert(false); }
};

/****************** PolicySet ************************/
template<typename P, typename UG = UserAndGroupContext<P>>
class PolicySet : public UG {
	/** Policies for all contexts */
	P contextMandatory;
	P contextDefault;

	P &getPolicyContext(ContextType type) {
		if (ContextType::MANDATORY == type)
			return contextMandatory;

		assert(ContextType::DEFAULT == type);
		return contextDefault;
	}

public:
	typedef P policy_type;
	/** Adds given item to policy
	 * \param[in] policy_type Policy type
	 * \param[in] policy_type_value Policy type value
	 * \param[in] item Item to add
	*/
	template<typename T>
	void addItem(const PolicyType policy_type,
				 const PolicyTypeValue policy_type_value,
				 T &item) {
		switch (policy_type) {
			case PolicyType::CONTEXT:
				getPolicyContext(policy_type_value.context).addItem(item);
				break;
			case PolicyType::USER:
				UG::addItemUser(policy_type_value.user, item);
				break;
			case PolicyType::GROUP:
				UG::addItemGroup(policy_type_value.group, item);
				break;
			case PolicyType::NONE:
				assert(false);
				break;
		}
	}

	const P &getRefPolicyContextMandatory() const { return contextMandatory; }
	const P &getRefPolicyContextDefault() const { return contextDefault; }
};

/****************** StorageBackendXML ************************/

/* Keep PolicyList, MatchItemTypes, and StorageBackendXML::StorageBackendXMLImpl::m_sets
 * in the same order (own, send, receive, access).
 */
using PolicyList = type_list::type_list<PolicyOwn, PolicySend, PolicyReceive, PolicyAccess>;
template <typename T> using PolicyIdx = type_list::IDX<T, PolicyList>;

class StorageBackendXML::StorageBackendXMLImpl {
public:
	std::tuple< PolicySet<PolicyOwn>,
				PolicySet<PolicySend>,
				PolicySet<PolicyReceive>,
				PolicySet<PolicyAccess, NoUserAndGroupContext> > m_sets;

	template <typename MatchItemType>
	auto getPolicySetFromMatchItem()
		-> typename std::tuple_element<MatchItemIdx<MatchItemType>::idx, decltype(m_sets)>::type &
	{
		return std::get<MatchItemIdx<MatchItemType>::idx>(m_sets);
	}

	template <typename PolicyType>
	auto getPolicySetFromPolicy() const
		-> const typename std::tuple_element<PolicyIdx<PolicyType>::idx, decltype(m_sets)>::type &
	{
		return std::get<PolicyIdx<PolicyType>::idx>(m_sets);
	}
};

template <typename T>
void StorageBackendXML::addItem(const ldp_xml_parser::PolicyType policy_type,
								const ldp_xml_parser::PolicyTypeValue policy_type_value,
								T &item) {
	tslog::log_debug("Add item: ", item, ", decision: ", item.getDecision(), "\n");
	pimpl->getPolicySetFromMatchItem<typename T::match_type>().addItem(policy_type, policy_type_value, item);
}

bool StorageBackendXML::init(const char *filename) {
	pimpl.reset(new StorageBackendXMLImpl);
	XmlParser parser(*this, filename);
	return parser.parsePolicyConfigFile() == 0;
}

template <typename T>
const T &StorageBackendXML::getPolicyContextMandatory() const {
	return pimpl->getPolicySetFromPolicy<T>().getRefPolicyContextMandatory();
}

template <typename T>
const T &StorageBackendXML::getPolicyContextDefault() const {
	return pimpl->getPolicySetFromPolicy<T>().getRefPolicyContextDefault();
}

template <typename T>
const std::map<uid_t, T> &StorageBackendXML::getPoliciesUser() const {
	return pimpl->getPolicySetFromPolicy<T>().getPoliciesUser();
}

template <typename T>
const std::map<gid_t, T> &StorageBackendXML::getPoliciesGroup() const {
	return pimpl->getPolicySetFromPolicy<T>().getPoliciesGroup();
}

StorageBackendXML::StorageBackendXML() : pimpl{new StorageBackendXMLImpl} {
}

StorageBackendXML::~StorageBackendXML() = default;

/* Explicit instantiation is needed for used public template methods defined in this file.
 */
#define T_INSTANTIATION(T) \
	template void StorageBackendXML::addItem(const PolicyType policy_type, const PolicyTypeValue policy_type_value, Item##T &item); \
	template const Policy##T &StorageBackendXML::getPolicyContextMandatory() const; \
	template const Policy##T &StorageBackendXML::getPolicyContextDefault() const;

T_INSTANTIATION(Own)
T_INSTANTIATION(Send)
T_INSTANTIATION(Receive)
T_INSTANTIATION(Access)

#undef T_INSTANTIATION

#define T_INSTANTIATION(T) \
	template const std::map<gid_t, Policy##T> &StorageBackendXML::getPoliciesGroup() const; \
	template const std::map<uid_t, Policy##T> &StorageBackendXML::getPoliciesUser() const;

T_INSTANTIATION(Own)
T_INSTANTIATION(Send)
T_INSTANTIATION(Receive)

#undef T_INSTANTIATION

} // namespace ldp_xml
