#include "internal/include/fb_generated.h"
#include "internal/naive_policy_checker.hpp"
#include "internal/policy.hpp"
#include "internal/serializer.hpp"
#include "internal/storage_backend_serialized.hpp"
#include "internal/storage_backend_serialized.hpp"
#include "internal/tslog.hpp"
#include "libdbuspolicy1-private.h"
#include <dbuspolicy1/libdbuspolicy1.h>
#include <getopt.h>
#include <iostream>
#include <string>
#include <sys/types.h>
#include <unistd.h>

using namespace ldp_xml_parser;
using namespace ldp_serialized;

enum class Choice {
	NONE,
	FB,
	XML,
	XMLplusFB,
	ALL,
};

bool measure(std::function<bool()> func, size_t count, const char *desc) {
	bool flag = true;
	clock_t begin = clock();
	for (size_t i = 0; i < count; i++)
		flag &= func();
	clock_t end = clock();

	std::cout << desc << ": " << static_cast<double>(end - begin)/CLOCKS_PER_SEC << std::endl;
	return flag;
}

bool run_xml(const char *conf_file) {
	return policy_checker_system().initDb(conf_file);
}

bool run_xml_plus_fb(const char *conf_file, bool verify) {
	Serializer serializer;
	size_t size;
	uint8_t *buff = serializer.serialize(conf_file, size);

	if (verify) {
		auto verifier = flatbuffers::Verifier(buff, size);
		if (!FB::VerifyFileBuffer(verifier)) {
			std::cout << "verification of serialized data: failed" << std::endl;
			return false;
		}
	}

	const FB::File *file = FB::GetFile(buff);

	StorageBackendSerialized storage;
	return storage.init(file);
}

bool run_fb(const char *conf_file, bool verify) {
	StorageBackendSerialized sbs;
	return sbs.init(conf_file, verify);
}

void run_tests(const char *conf_file, const char *conf_bin, size_t c, Choice ch, bool verify) {
	if (ch == Choice::ALL || ch == Choice::XML) {
		if (!measure([&conf_file, c]() { return run_xml(conf_file); }, c, "XML")) {
			std::cout << "ERROR" << std::endl;
		}
	}
	if (ch == Choice::ALL || ch == Choice::FB) {
		if (!measure([&conf_bin, c, verify]() { return run_fb(conf_bin, verify); }, c, "FB")) {
			std::cout << "ERROR" << std::endl;
		}
	}
	if (ch == Choice::ALL || ch == Choice::XMLplusFB)
		if (!measure([&conf_file, c, verify]() { return run_xml_plus_fb(conf_file, verify); }, c, "FB after XML")) {
			std::cout << "ERROR" << std::endl;
		}
}

void print_help(const char *name) {
	std::cout << std::endl;
	std::cout << "usage: " << name << " {-f <config_bin>|-x|-d|-a <config_bin>} {--system|--session|-c <config_xml>} <count>" << std::endl;
	std::cout << std::endl;
	std::cout << "       -f <config_bin> - Flatbuffers" << std::endl;
	std::cout << "       -x              - XML" << std::endl;
	std::cout << "       -d              - FB after XML" << std::endl;
	std::cout << "       -a <config_bin> - All tests" << std::endl;
	std::cout << "       -v              - Verify" << std::endl;
	std::cout << std::endl;
}

static const struct option options[] {
	{"system", no_argument, 0, 0},
	{"session", no_argument, 0, 0}
};

int main(int argc, char *argv[])
{
	int c;
	std::string input_filename = system_bus_conf_file_primary();
	std::string binary_file = "";
	size_t count = 100;
	Choice choice = Choice::NONE;
	bool verify = false;

	while (1) {
		int option_index;
		c = getopt_long(argc, argv, "f:xda:c:v", options, &option_index);
		if (c == -1)
			break;
		switch(c) {
		case 0:
			if (option_index == 1)
				input_filename = session_bus_conf_file_primary();
			break;
		case 'a':
			if (choice != Choice::NONE) {
				print_help(argv[0]);
				return -1;
			}
			choice = Choice::ALL;
			binary_file = optarg;
			break;
		case 'f':
			if (choice != Choice::NONE) {
				print_help(argv[0]);
				return -1;
			}
			choice = Choice::FB;
			binary_file = optarg;
			break;
		case 'x':
			if (choice != Choice::NONE) {
				print_help(argv[0]);
				return -1;
			}
			choice = Choice::XML;
			break;
		case 'd':
			if (choice != Choice::NONE) {
				print_help(argv[0]);
				return -1;
			}
			choice = Choice::XMLplusFB;
			break;
		case 'c':
			input_filename = optarg;
			break;
		case 'v':
			verify = true;
			break;
		}
	}

	if (optind < argc) {
		count = std::stoi(argv[optind]);
	} else {
		print_help(argv[0]);
		return 1;
	}

	tslog::init();
	run_tests(input_filename.c_str(), binary_file.c_str(), count, choice, verify);

	return 0;
}
